//
// merges many files into one file
//
// Written by Maxwell Sayles.  Copyright (C) 2000.
//

#include "FileMerger.hpp"
#include "util.hpp"

//
// creation
//
FileMerger::FileMerger (char* outFileName, char** list_inFileNames, int number_inFiles, int bufferSizeInKB, bool deleteFiles)
{
	this->deleteFiles						= deleteFiles;

	this->outFileName						= outFileName;
	this->outFileTitle					= createFileName_from_fullPath (outFileName);
	this->totalSize							= 0;

	this->list_inFileNames			= list_inFileNames;
	this->number_inFiles				= number_inFiles;
	this->list_inFileSizes			= new INT64 [this->number_inFiles];

	// get each file's size and tally
	{
		int index = 0;
		while (index < this->number_inFiles)
		{
			// get file size
			HANDLE file = CreateFile (this->list_inFileNames [index], 0, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);
			DWORD hiSize = 0;
			DWORD loSize = GetFileSize (file, &hiSize);
			CloseHandle (file);
			INT64 fileSize = (INT64 (hiSize) << 32) + INT64 (loSize);

			this->list_inFileSizes [index] = fileSize;
			this->totalSize += fileSize;

			++ index;
		}
	}

	this->bufferSize						= bufferSizeInKB * 1024;
	this->buffer								= new BYTE [this->bufferSize];

	this->index_fileList				= -1;

	this->bytesReadOfInput			= 0;
	this->bytesToReadOfInput		= 0;
	this->bytesWrittenOfOutput	= 0;

	this->inFile								= 0;
	this->outFile = CreateFile (this->outFileName, GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_ARCHIVE, NULL);

	this->error									= false;
	this->errorCode							= 0;
	this->completed							= false;

	// check for error of creation file
	if (this->outFile == INVALID_HANDLE_VALUE)
	{
		this->error = true;
		this->errorCode = GetLastError ();
	}
}



//
// destruction
//
FileMerger::~FileMerger ()
{
	if (this->inFile != 0)
	{
		CloseHandle (this->inFile);
	}
	CloseHandle (this->outFile);
	delete[] this->outFileTitle;

	delete[] this->list_inFileSizes;
	delete[] this->buffer;
}


//
// is finished?
//
bool
FileMerger::isFinished ()
{
	return (this->error || this->completed);
}


//
// copy one buffer worth of data
//
void
FileMerger::copyOneBuffer ()
{
	if (isFinished ()) return;

	// did we finish reading the last file?
	if (this->bytesReadOfInput == this->bytesToReadOfInput)
	{
		++ this->index_fileList;

		this->bytesReadOfInput = 0;
		this->bytesToReadOfInput = this->list_inFileSizes [this->index_fileList];

		this->inFile = CreateFile (this->list_inFileNames [this->index_fileList], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (this->inFile == INVALID_HANDLE_VALUE)
		{
			this->error = true;
			this->errorCode = GetLastError ();
			return;
		}
	}

	// read a block from input file
	DWORD bytesRead = 0;
	DWORD bytesToRead = (DWORD)(this->bufferSize);
	if ((this->bytesToReadOfInput - this->bytesReadOfInput) < bytesToRead)
	{
		bytesToRead = (DWORD)(this->bytesToReadOfInput - this->bytesReadOfInput);
	}
	BOOL readResult = ReadFile (this->inFile, (LPVOID)(this->buffer), bytesToRead, &bytesRead, NULL);

	if (readResult == 0)
	{
		// signal an error
		this->error = true;
		this->errorCode = GetLastError ();
	}
	else
	{
		// write output
		DWORD bytesWritten = 0;
		BOOL writeResult = WriteFile (this->outFile, (LPCVOID)(this->buffer), bytesRead, &bytesWritten, NULL);

		if ((writeResult == 0) ||
				(bytesWritten != bytesRead))
		{
			this->error = true;
			this->errorCode = GetLastError ();
		}
		else
		{

			this->bytesReadOfInput += bytesRead;
			this->bytesWrittenOfOutput += bytesRead;

			// read entire file?
			if (this->bytesReadOfInput == this->bytesToReadOfInput)
			{
				// close file
				CloseHandle (this->inFile);
				this->inFile = 0;

				// delete file?
				if (this->deleteFiles)
				{
					DeleteFile (this->list_inFileNames [this->index_fileList]);
				}

				// are we finished?
				if (this->bytesWrittenOfOutput == this->totalSize)
				{
					this->completed = true;
				}

			} // read entire file?

		} // write check

	} // read check

} // copyOneBuffer ()
