//
// file merger dialog
//
// Written by Maxwell Sayles.  Copyright (C) 2000.
//

#include <windows.h>
#include <stdio.h>
#include "resource.h"
#include "FileMerger_Dialog.hpp"
#include "FileMerger.hpp"
#include "Merge_Dialog.hpp"
#include "util.hpp"


static const char* windowTitle = "File Merger";

//
// dispatching for About_Dialog
//
BOOL CALLBACK About_DialogProc( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{

	switch (uMsg)
	{
		case WM_INITDIALOG:
			{
				char buildString[1000] = "Built ";
				char* date = date_mmmddyyyy_toWritten (__DATE__);
				strcat (buildString, date);
				delete[] date;
				strcat (buildString, ".");
				SendDlgItemMessage (hWndDlg, IDC_BUILD_DATE, WM_SETTEXT, 0, (LPARAM)(buildString));
			}
		return TRUE;

		case WM_CLOSE:
			EndDialog (hWndDlg, 0);
		return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDCANCEL:
					PostMessage (hWndDlg, WM_CLOSE, 0, 0);
				return TRUE;
			}
		return FALSE;
	}

	return FALSE;
}


//
// dispatching
//
BOOL CALLBACK
FileMerger_Dialog::DialogProc( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	FileMerger_Dialog* dialog = (FileMerger_Dialog*)(GetWindowLong (hWndDlg, GWL_USERDATA));

	switch (uMsg)
	{
		case WM_INITDIALOG:
			dialog = (FileMerger_Dialog*)(lParam);
			SetWindowLong (hWndDlg, GWL_USERDATA, (LONG)(dialog));
			dialog->hWnd = hWndDlg;
			dialog->initDialog ();
		return TRUE;

		case WM_CLOSE:
			dialog->close ();
			EndDialog (hWndDlg, 0);
		return TRUE;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDC_MERGE:
					dialog->merge();
				return TRUE;

				case IDC_MOVE_UP:
					dialog->moveUp();
				return TRUE;

				case IDC_MOVE_DOWN:
					dialog->moveDown();
				return TRUE;

				case IDC_ADD_FILES:
					dialog->addFiles();
				return TRUE;

				case IDC_REMOVE_FILES:
					dialog->removeFiles();
				return TRUE;

				case IDC_ABOUT:
				{
					HINSTANCE hInstance = (HINSTANCE)(GetWindowLong (hWndDlg, GWL_HINSTANCE));
					DialogBox (hInstance, MAKEINTRESOURCE (IDD_ABOUT), hWndDlg, (DLGPROC)(About_DialogProc));
				}
				return TRUE;

				case IDC_DESTINATION:
					dialog->selectDestinationFile();
				return TRUE;

				case IDC_CLOSE:
					PostMessage (hWndDlg, WM_CLOSE, 0, 0);
				return TRUE;
			}
		return FALSE;

		case WM_ENTERIDLE:
			dialog->enterIdle ();
		return TRUE;

		case WM_TIMER:
			dialog->onTimer ();
		return TRUE;

	}

	return FALSE;
}


//
// constructor
//
FileMerger_Dialog::FileMerger_Dialog (char* destinationFileName, bool deleteAfterProcessed)
	:	hWnd (0),
		mergeDialog (0),
		merger (0),
		titleIconIndex (0)
{
	this->destinationFileName = new char [strlen (destinationFileName) + 1];
	strcpy (this->destinationFileName, destinationFileName);

	this->deleteAfterProcessed = deleteAfterProcessed;

	HINSTANCE hInstance = (HINSTANCE)(GetWindowLong (this->hWnd, GWL_HINSTANCE));
	this->hBmpOpenFile = LoadBitmap (hInstance, MAKEINTRESOURCE (IDB_OPEN_FILE));
	this->hIconTitle [0] = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_GLOW00));
	this->hIconTitle [1] = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_GLOW15));
	this->hIconTitle [2] = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_GLOW30));
	this->hIconTitle [3] = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_GLOW45));
	this->hIconTitle [4] = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_GLOW60));
	this->hIconTitle [5] = LoadIcon (hInstance, MAKEINTRESOURCE (IDI_GLOW75));
}


//
// destructor
//
FileMerger_Dialog::~FileMerger_Dialog ()
{
	DestroyIcon (this->hIconTitle [0]);
	DestroyIcon (this->hIconTitle [1]);
	DestroyIcon (this->hIconTitle [2]);
	DestroyIcon (this->hIconTitle [3]);
	DestroyIcon (this->hIconTitle [4]);
	DestroyIcon (this->hIconTitle [5]);

	DeleteObject (this->hBmpOpenFile);
	delete[] this->destinationFileName;
}


//
// on timer
//
// little spinny icon thing
//
void
FileMerger_Dialog::onTimer ()
{
	++ this->titleIconIndex;
	if (this->titleIconIndex == 6)
	{
		this->titleIconIndex = 0;
	}

	SendMessage (this->hWnd, WM_SETICON, ICON_SMALL, (LPARAM)(this->hIconTitle [this->titleIconIndex]));
}


//
// on close
//
void
FileMerger_Dialog::close ()
{
	// kill the timer
	KillTimer (this->hWnd, 1);

	// get current destination file name
	delete[] this->destinationFileName;

	LONG length = SendDlgItemMessage (this->hWnd, IDC_DESTINATION_FILE, WM_GETTEXTLENGTH, 0, 0);
	this->destinationFileName = new char [length + 1];
	SendDlgItemMessage (this->hWnd, IDC_DESTINATION_FILE, WM_GETTEXT, (WPARAM)(length + 1), (LPARAM)(this->destinationFileName));

	// get status of 'delete after processed' check box
	LONG status = SendDlgItemMessage (this->hWnd, IDC_CHECK_DELETE_FILES, BM_GETCHECK, 0, 0);
	this->deleteAfterProcessed = (status == BST_CHECKED);	
}


//
// init dialog
//
void
FileMerger_Dialog::initDialog ()
{

	SendDlgItemMessage (this->hWnd, IDC_BUFFER_SIZE, WM_SETTEXT, 0, (LPARAM)("4096"));
	SendDlgItemMessage (this->hWnd, IDC_DESTINATION_FILE, WM_SETTEXT, 0, (LPARAM)(this->destinationFileName));
	SendDlgItemMessage (this->hWnd, IDC_DESTINATION, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)(this->hBmpOpenFile));
	SendMessage (this->hWnd, WM_SETICON, ICON_BIG, (LPARAM)(this->hIconTitle [0]));
	SendMessage (this->hWnd, WM_SETICON, ICON_SMALL, (LPARAM)(this->hIconTitle [0]));

	if (this->deleteAfterProcessed)
	{
		SendDlgItemMessage (this->hWnd, IDC_CHECK_DELETE_FILES, BM_SETCHECK, BST_CHECKED, 0);
	}

	// create timer for icon updates
	SetTimer (this->hWnd, 1, 100, NULL);
}


//
// perform merge
//
// lower the priority of the thread
// create the file merger
// merge the files (updating user of progress)
// destroy file merger
// restore priority
//
void
FileMerger_Dialog::doMerge ()
{
	// set the thread priority to one below normal, so most apps will still run smootly
	HANDLE thread = GetCurrentThread ();
	int oldThreadPriority = GetThreadPriority (thread);
	SetThreadPriority (thread, THREAD_PRIORITY_BELOW_NORMAL);

	// kill the timer
	KillTimer (this->hWnd, 1);

	// get filename
	HWND hWndDestinationFile = GetDlgItem (this->hWnd, IDC_DESTINATION_FILE);
	int length = (int)SendMessage (hWndDestinationFile, WM_GETTEXTLENGTH, 0, 0);
	char* outFileName = new char [length + 1];
	SendMessage (hWndDestinationFile, WM_GETTEXT, (WPARAM)(length + 1), (LPARAM)(outFileName));

	// get number of in files and each in file

	// get count
	HWND hWndFileList = GetDlgItem (this->hWnd, IDC_FILE_LIST);
	int number_inFiles = (int)SendMessage (hWndFileList, LB_GETCOUNT, 0, 0);

	// allocate array
	char** list_inFileNames = new char* [number_inFiles];

	// read in each file name
	int index = 0;
	while (index < number_inFiles)
	{
		int length = (int)SendMessage (hWndFileList, LB_GETTEXTLEN, (WPARAM)(index), 0);
		list_inFileNames [index] = new char [length + 1];
		SendMessage (hWndFileList, LB_GETTEXT, (WPARAM)(index), (LPARAM)(list_inFileNames [index]));

		++ index;
	}

	// get buffer size
	char bufferSizeString [13];
	GetDlgItemText (this->hWnd, IDC_BUFFER_SIZE, bufferSizeString, 12);
	int bufferSize = atoi (bufferSizeString);

	// get delete files check state
	LONG deleteFilesCheckState = SendDlgItemMessage (this->hWnd, IDC_CHECK_DELETE_FILES, BM_GETCHECK, 0, 0);

	// create file merger that will run when idle
	this->merger = new FileMerger (outFileName, list_inFileNames, number_inFiles, bufferSize, (deleteFilesCheckState == BST_CHECKED));
	this->currentFileIndex = this->merger->inputIndexInFileList ();

	// create dialog box
	HINSTANCE hInstance = (HINSTANCE)(GetWindowLong (this->hWnd, GWL_HINSTANCE));
	this->mergeDialog = new Merge_Dialog ();
	DialogBoxParam (hInstance, MAKEINTRESOURCE (IDD_MERGE), this->hWnd, (DLGPROC)Merge_Dialog::DialogProc, (LPARAM)(this->mergeDialog));
	delete this->mergeDialog;
	this->mergeDialog = 0;

	// restore window title
	SetWindowText (this->hWnd, windowTitle);

	// report status of merge
	int messageResult = IDOK;
	if (this->merger->completed)
	{
		MessageBox (this->hWnd, "Merge successful.", "Success", MB_OK);
	}
	else if (this->merger->error)
	{
		char errorMessage [1025];
		FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM, 0, this->merger->errorCode, 0, errorMessage, 1024, 0);
		char output [2048];
		sprintf (output, "Error:\n%s\nDo you want to keep the output file?", errorMessage);
		messageResult = MessageBox (this->hWnd, output, "Keep output file?", MB_YESNO | MB_ICONEXCLAMATION);
	}
	else
	{
		messageResult = MessageBox (this->hWnd, "Merge canceled.\nDo you want to keep the output file?", "Keep output file?", MB_YESNO);
	}

	// destroy merge and make sure all files are closed
	delete this->merger;
	this->merger = 0;


	// delete output file?
	if (messageResult == IDNO)
	{
		// delete it
		DeleteFile (outFileName);
	}



	// shut everything down
	index = 0;
	while (index < number_inFiles)
	{
		delete[] list_inFileNames [index];
		++ index;
	}

	delete[] list_inFileNames;
	delete[] outFileName;

	// create timer for icon updates
	SetTimer (this->hWnd, 1, 100, NULL);

	// restore thread priority
	SetThreadPriority (thread, oldThreadPriority);
}


//
// merge pressed
// get the number of files in list
// if zero then return message
// else perform merge
//
void
FileMerger_Dialog::merge ()
{
	// get count of input files
	int fileCount = (int)(SendDlgItemMessage (this->hWnd, IDC_FILE_LIST, LB_GETCOUNT, 0, 0));

	if (fileCount > 0)
	{
		this->doMerge ();
	}
	else
	{
		MessageBox (this->hWnd, "You must select files to merge.", "Error", MB_OK | MB_ICONEXCLAMATION);
	}
}


//
// move up button was pressed
// start the first item
// if both it is not selected and the item below is selected then swap them
// repeat on the below item
//
void
FileMerger_Dialog::moveUp ()
{
	HWND hWndFileList = GetDlgItem (this->hWnd, IDC_FILE_LIST);

	// get number of items in list
	int itemCount = (int)SendMessage (hWndFileList, LB_GETCOUNT, 0, 0);

	int index = 0;
	LRESULT isCurrentSelected = SendMessage (hWndFileList, LB_GETSEL, 0, 0);
	while (index < (itemCount - 1))
	{
		LRESULT isNextSelected = SendMessage (hWndFileList, LB_GETSEL, index + 1, 0);

		if ((isCurrentSelected == 0) &&
				(isNextSelected > 0))
		{
			// swap them

			int	length			= (int)SendMessage (hWndFileList, LB_GETTEXTLEN, index, 0);
			char* string		= new char[length + 1];
			SendMessage (hWndFileList, LB_GETTEXT, index, (LPARAM)string);
			SendMessage (hWndFileList, LB_DELETESTRING, index, 0);
			SendMessage (hWndFileList, LB_INSERTSTRING, index + 1, (LPARAM)string);
			delete string;

		}
		else
		{
			isCurrentSelected = isNextSelected;
		}

		++ index;
	}
}


//
// move down button was pressed
// start at the last item in the list, 
// if both it is not selected and the item above it is selected then swap them
// repeat on the above item
//
void
FileMerger_Dialog::moveDown ()
{
	HWND hWndFileList = GetDlgItem (this->hWnd, IDC_FILE_LIST);

	// get number of items in list
	int itemCount = (int)SendMessage (hWndFileList, LB_GETCOUNT, 0, 0);

	int index = itemCount - 1;
	LRESULT isCurrentSelected = SendMessage (hWndFileList, LB_GETSEL, index, 0);
	while (index > 0)
	{
		LRESULT isNextSelected = SendMessage (hWndFileList, LB_GETSEL, index - 1, 0);

		if ((isCurrentSelected == 0) &&
				(isNextSelected > 0))
		{
			// swap them

			int	length			= (int)SendMessage (hWndFileList, LB_GETTEXTLEN, index, 0);
			char* string		= new char[length + 1];
			SendMessage (hWndFileList, LB_GETTEXT, index, (LPARAM)string);
			SendMessage (hWndFileList, LB_DELETESTRING, index, 0);
			SendMessage (hWndFileList, LB_INSERTSTRING, index - 1, (LPARAM)string);
			delete string;

		}
		else
		{
			isCurrentSelected = isNextSelected;
		}

		-- index;
	}
}


//
// add files
// pop up browse window, allow multiple select of files, add each file to the list
//
void
FileMerger_Dialog::addFiles ()
{
	char fileName[1024] = "";
	OPENFILENAME openFileName;
	ZeroMemory (&openFileName, sizeof (OPENFILENAME));
	openFileName.lStructSize	= sizeof (OPENFILENAME);
	openFileName.hwndOwner		= this->hWnd;
	openFileName.lpstrFilter	= "All Files\0*.*\0";
	openFileName.lpstrFile		= fileName;
	openFileName.nMaxFile			= sizeof (fileName);
	openFileName.lpstrTitle		= "Add File(s)";
	openFileName.Flags				= OFN_ALLOWMULTISELECT | OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_READONLY;
	if (GetOpenFileName (&openFileName) != 0)
	{
		// process selection
		HWND hWndFileList = GetDlgItem (this->hWnd, IDC_FILE_LIST);
		
		// determine single select/multiple selection
		if (openFileName.nFileOffset > strlen (fileName))
		{
			char* namePointer = fileName + openFileName.nFileOffset;
			char name[1024] = "";
			bool looping = true;
			while (looping)
			{
				// build file name
				strcpy (name, fileName);
				strcat (name, "\\");
				strcat (name, namePointer);

				// add it to list box
				SendMessage (hWndFileList, LB_ADDSTRING, 0, (LPARAM)(name));

				// next string
				while (*namePointer != '\0')
				{
					++ namePointer;
				}
				++ namePointer;

				// test if string is a null character (end of list)
				if (*namePointer == '\0')
				{
					looping = false;
				}
			}
		}
		else
		{
			SendMessage (hWndFileList, LB_ADDSTRING, 0, (LPARAM)(fileName));
		}
	}
}


//
// remove all selected files from list
// scroll each item, if the item is selected then remove it
//
void
FileMerger_Dialog::removeFiles ()
{
	HWND hWndFileList = GetDlgItem (this->hWnd, IDC_FILE_LIST);
	int count = (int)SendMessage (hWndFileList, LB_GETCOUNT, 0, 0);

	int index = 0;
	while (index < count)
	{
		int isSelected = (int)SendMessage (hWndFileList, LB_GETSEL, (WPARAM)(index), 0);
		if (isSelected > 0)
		{
			SendMessage (hWndFileList, LB_DELETESTRING, (WPARAM)(index), 0);
			-- count;
		}
		else
		{
			++ index;
		}
	}
}


//
// select destination file
//
void
FileMerger_Dialog::selectDestinationFile ()
{
	char fileName[1024] = "";
	OPENFILENAME openFileName;
	ZeroMemory (&openFileName, sizeof (OPENFILENAME));
	openFileName.lStructSize	= sizeof (OPENFILENAME);
	openFileName.hwndOwner		= this->hWnd;
	openFileName.lpstrFilter	= "All Files\0*.*\0";
	openFileName.lpstrFile		= fileName;
	openFileName.nMaxFile			= sizeof (fileName);
	openFileName.lpstrTitle		= "Select Destination File";
	openFileName.Flags				= OFN_EXPLORER;
	if (GetSaveFileName (&openFileName) != 0)
	{
		// process selection
		HWND hWndDestinationFile = GetDlgItem (this->hWnd, IDC_DESTINATION_FILE);
		SetWindowText (hWndDestinationFile, fileName);
	}
}



//
// enter idle
//
void
FileMerger_Dialog::enterIdle ()
{
	if (this->mergeDialog != 0)
	{
		if (!this->mergeDialog->isPaused)
		{
			if (this->currentFileIndex == -1)
			{
				mergeDialog->setUI_outputFileName (this->merger->currentOutputFile ());
			}

			if (!this->merger->isFinished ())
			{
				this->merger->copyOneBuffer ();
				this->mergeDialog->setUI_inputPercentage (this->merger->inputPercentComplete ());
				this->mergeDialog->setUI_outputPercentage (this->merger->ouputPercentComplete ());

				// adjust window title to reflect percentage
				char string [1024];
				sprintf (string, "%2.0f%% [%s] - %s", this->merger->ouputPercentComplete (), this->merger->currentOutputFileTitle(), windowTitle);
				SetWindowText (this->hWnd, string);
				
			}
			else
			{
				this->mergeDialog->close ();
			}
			
			if (this->currentFileIndex != this->merger->inputIndexInFileList ())
			{
				this->currentFileIndex = this->merger->inputIndexInFileList ();
				mergeDialog->setUI_inputFileName (this->merger->currentInputFile ());
			}

			PostMessage (this->hWnd, WM_TIMER, 0, 0);
		}
	}
}
